@@ -1,9 +1,16 @@
+from random import Random
+
 from eth2spec.test.context import (
-    spec_state_test, expect_assertion_error, always_bls, with_all_phases
+    spec_state_test, expect_assertion_error, always_bls, with_all_phases,
+    with_custom_state, spec_test, single_phase,
+    low_balances, misc_balances,
 )
 from eth2spec.test.helpers.attestations import sign_indexed_attestation
-from eth2spec.test.helpers.attester_slashings import get_valid_attester_slashing, \
-    get_indexed_attestation_participants, get_attestation_2_data, get_attestation_1_data
+from eth2spec.test.helpers.attester_slashings import (
+    get_valid_attester_slashing, get_valid_attester_slashing_by_indices,
+    get_indexed_attestation_participants, get_attestation_2_data, get_attestation_1_data,
+)
+from eth2spec.test.helpers.proposer_slashings import get_min_slashing_penalty_quotient
 from eth2spec.test.helpers.state import (
     get_balance,
     next_epoch_via_block,
@@ -31,15 +38,19 @@ def run_attester_slashing_processing(spec, state, attester_slashing, valid=True)
 
     proposer_index = spec.get_beacon_proposer_index(state)
     pre_proposer_balance = get_balance(state, proposer_index)
-    pre_slashings = {slashed_index: get_balance(state, slashed_index) for slashed_index in slashed_indices}
+    pre_slashing_balances = {slashed_index: get_balance(state, slashed_index) for slashed_index in slashed_indices}
+    pre_slashing_effectives = {
+        slashed_index: state.validators[slashed_index].effective_balance
+        for slashed_index in slashed_indices
+    }
     pre_withdrawalable_epochs = {
         slashed_index: state.validators[slashed_index].withdrawable_epoch
         for slashed_index in slashed_indices
     }
 
     total_proposer_rewards = sum(
-        balance // spec.WHISTLEBLOWER_REWARD_QUOTIENT
-        for balance in pre_slashings.values()
+        effective_balance // spec.WHISTLEBLOWER_REWARD_QUOTIENT
+        for effective_balance in pre_slashing_effectives.values()
     )
 
     # Process slashing
@@ -60,7 +71,7 @@ def run_attester_slashing_processing(spec, state, attester_slashing, valid=True)
             assert slashed_validator.withdrawable_epoch == expected_withdrawable_epoch
         else:
             assert slashed_validator.withdrawable_epoch < spec.FAR_FUTURE_EPOCH
-        assert get_balance(state, slashed_index) < pre_slashings[slashed_index]
+        assert get_balance(state, slashed_index) < pre_slashing_balances[slashed_index]
 
     if proposer_index not in slashed_indices:
         # gained whistleblower reward
@@ -70,7 +81,7 @@ def run_attester_slashing_processing(spec, state, attester_slashing, valid=True)
         expected_balance = (
             pre_proposer_balance
             + total_proposer_rewards
-            - pre_slashings[proposer_index] // spec.MIN_SLASHING_PENALTY_QUOTIENT
+            - pre_slashing_effectives[proposer_index] // get_min_slashing_penalty_quotient(spec)
         )
 
         assert get_balance(state, proposer_index) == expected_balance
@@ -117,6 +128,75 @@ def test_success_already_exited_recent(spec, state):
     yield from run_attester_slashing_processing(spec, state, attester_slashing)
 
 
+@with_all_phases
+@spec_state_test
+@always_bls
+def test_success_proposer_index_slashed(spec, state):
+    # Transition past genesis slot because generally doesn't have a proposer
+    next_epoch_via_block(spec, state)
+
+    proposer_index = spec.get_beacon_proposer_index(state)
+    attester_slashing = get_valid_attester_slashing_by_indices(
+        spec, state,
+        [proposer_index],
+        signed_1=True, signed_2=True,
+    )
+
+    yield from run_attester_slashing_processing(spec, state, attester_slashing)
+
+
+@with_all_phases
+@spec_state_test
+def test_success_attestation_from_future(spec, state):
+    # Transition state to future to enable generation of a "future" attestation
+    future_state = state.copy()
+    next_epoch_via_block(spec, future_state)
+    # Generate slashing using the future state
+    attester_slashing = get_valid_attester_slashing(
+        spec, future_state,
+        slot=state.slot + 5,  # Slot is in the future wrt `state`
+        signed_1=True, signed_2=True
+    )
+
+    yield from run_attester_slashing_processing(spec, state, attester_slashing)
+
+
+@with_all_phases
+@with_custom_state(balances_fn=low_balances, threshold_fn=lambda spec: spec.config.EJECTION_BALANCE)
+@spec_test
+@single_phase
+def test_success_low_balances(spec, state):
+    attester_slashing = get_valid_attester_slashing(spec, state, signed_1=True, signed_2=True)
+
+    yield from run_attester_slashing_processing(spec, state, attester_slashing)
+
+
+@with_all_phases
+@with_custom_state(balances_fn=misc_balances, threshold_fn=lambda spec: spec.config.EJECTION_BALANCE)
+@spec_test
+@single_phase
+def test_success_misc_balances(spec, state):
+    attester_slashing = get_valid_attester_slashing(spec, state, signed_1=True, signed_2=True)
+
+    yield from run_attester_slashing_processing(spec, state, attester_slashing)
+
+
+@with_all_phases
+@with_custom_state(balances_fn=misc_balances, threshold_fn=lambda spec: spec.config.EJECTION_BALANCE)
+@spec_test
+@single_phase
+def test_success_with_effective_balance_disparity(spec, state):
+    # Jitter balances to be different from effective balances
+    rng = Random(12345)
+    for i in range(len(state.balances)):
+        pre = int(state.balances[i])
+        state.balances[i] += rng.randrange(max(pre - 5000, 0), pre + 5000)
+
+    attester_slashing = get_valid_attester_slashing(spec, state, signed_1=True, signed_2=True)
+
+    yield from run_attester_slashing_processing(spec, state, attester_slashing)
+
+
 @with_all_phases
 @spec_state_test
 @always_bls
@@ -226,7 +306,7 @@ def test_att1_empty_indices(spec, state):
     attester_slashing = get_valid_attester_slashing(spec, state, signed_1=False, signed_2=True)
 
     attester_slashing.attestation_1.attesting_indices = []
-    attester_slashing.attestation_1.signature = spec.bls.Z2_SIGNATURE
+    attester_slashing.attestation_1.signature = spec.bls.G2_POINT_AT_INFINITY
 
     yield from run_attester_slashing_processing(spec, state, attester_slashing, False)
 
@@ -238,7 +318,7 @@ def test_att2_empty_indices(spec, state):
     attester_slashing = get_valid_attester_slashing(spec, state, signed_1=True, signed_2=False)
 
     attester_slashing.attestation_2.attesting_indices = []
-    attester_slashing.attestation_2.signature = spec.bls.Z2_SIGNATURE
+    attester_slashing.attestation_2.signature = spec.bls.G2_POINT_AT_INFINITY
 
     yield from run_attester_slashing_processing(spec, state, attester_slashing, False)
 
@@ -250,10 +330,10 @@ def test_all_empty_indices(spec, state):
     attester_slashing = get_valid_attester_slashing(spec, state, signed_1=False, signed_2=False)
 
     attester_slashing.attestation_1.attesting_indices = []
-    attester_slashing.attestation_1.signature = spec.bls.Z2_SIGNATURE
+    attester_slashing.attestation_1.signature = spec.bls.G2_POINT_AT_INFINITY
 
     attester_slashing.attestation_2.attesting_indices = []
-    attester_slashing.attestation_2.signature = spec.bls.Z2_SIGNATURE
+    attester_slashing.attestation_2.signature = spec.bls.G2_POINT_AT_INFINITY
 
     yield from run_attester_slashing_processing(spec, state, attester_slashing, False)
 
