@@ -19,22 +19,27 @@ from eth2spec.test.helpers.attester_slashings import (
 from eth2spec.test.helpers.proposer_slashings import get_valid_proposer_slashing, check_proposer_slashing_effect
 from eth2spec.test.helpers.attestations import get_valid_attestation
 from eth2spec.test.helpers.deposits import prepare_state_and_deposit
+from eth2spec.test.helpers.execution_payload import build_empty_execution_payload
 from eth2spec.test.helpers.voluntary_exits import prepare_signed_exits
-from eth2spec.test.helpers.shard_transitions import get_shard_transition_of_committee
 from eth2spec.test.helpers.multi_operations import (
     run_slash_and_exit,
     run_test_full_random_operations,
 )
-
+from eth2spec.test.helpers.sync_committee import (
+    compute_committee_indices,
+    compute_sync_committee_participant_reward_and_penalty,
+)
+from eth2spec.test.helpers.constants import PHASE0, MINIMAL
 from eth2spec.test.context import (
-    PHASE0, PHASE1, MINIMAL,
     spec_test, spec_state_test, dump_skipping_message,
     with_phases, with_all_phases, single_phase,
     expect_assertion_error, always_bls,
     disable_process_reveal_deadlines,
-    with_configs,
+    with_presets,
     with_custom_state,
     large_validator_set,
+    is_post_altair,
+    is_post_merge,
 )
 
 
@@ -99,10 +104,10 @@ def test_empty_block_transition(spec, state):
 
 
 @with_all_phases
-@with_configs([MINIMAL],
+@with_presets([MINIMAL],
               reason="mainnet config leads to larger validator set than limit of public/private keys pre-generated")
 @spec_test
-@with_custom_state(balances_fn=large_validator_set, threshold_fn=lambda spec: spec.EJECTION_BALANCE)
+@with_custom_state(balances_fn=large_validator_set, threshold_fn=lambda spec: spec.config.EJECTION_BALANCE)
 @single_phase
 def test_empty_block_transition_large_validator_set(spec, state):
     pre_slot = state.slot
@@ -138,11 +143,16 @@ def process_and_sign_block_without_header_validations(spec, state, block):
         state_root=spec.Bytes32(),
         body_root=block.body.hash_tree_root(),
     )
+    if is_post_merge(spec):
+        if spec.is_execution_enabled(state, block.body):
+            spec.process_execution_payload(state, block.body.execution_payload, spec.EXECUTION_ENGINE)
 
     # Perform rest of process_block transitions
     spec.process_randao(state, block.body)
     spec.process_eth1_data(state, block.body)
     spec.process_operations(state, block.body)
+    if is_post_altair(spec):
+        spec.process_sync_aggregate(state, block.body.sync_aggregate)
 
     # Insert post-state rot
     block.state_root = state.hash_tree_root()
@@ -185,6 +195,9 @@ def test_parent_from_same_slot(spec, state):
     child_block = parent_block.copy()
     child_block.parent_root = state.latest_block_header.hash_tree_root()
 
+    if is_post_merge(spec):
+        child_block.body.execution_payload = build_empty_execution_payload(spec, state)
+
     # Show that normal path through transition fails
     failed_state = state.copy()
     expect_assertion_error(
@@ -327,10 +340,10 @@ def test_empty_epoch_transition(spec, state):
 
 
 @with_all_phases
-@with_configs([MINIMAL],
+@with_presets([MINIMAL],
               reason="mainnet config leads to larger validator set than limit of public/private keys pre-generated")
 @spec_test
-@with_custom_state(balances_fn=large_validator_set, threshold_fn=lambda spec: spec.EJECTION_BALANCE)
+@with_custom_state(balances_fn=large_validator_set, threshold_fn=lambda spec: spec.config.EJECTION_BALANCE)
 @single_phase
 def test_empty_epoch_transition_large_validator_set(spec, state):
     pre_slot = state.slot
@@ -417,7 +430,7 @@ def test_proposer_slashing(spec, state):
     yield 'blocks', [signed_block]
     yield 'post', state
 
-    check_proposer_slashing_effect(spec, pre_state, state, slashed_index)
+    check_proposer_slashing_effect(spec, pre_state, state, slashed_index, block)
 
 
 @with_all_phases
@@ -492,7 +505,7 @@ def test_multiple_different_proposer_slashings_same_block(spec, state):
 
     for proposer_slashing in proposer_slashings:
         slashed_index = proposer_slashing.signed_header_1.message.proposer_index
-        check_proposer_slashing_effect(spec, pre_state, state, slashed_index)
+        check_proposer_slashing_effect(spec, pre_state, state, slashed_index, block)
 
 
 def check_attester_slashing_effect(spec, pre_state, state, slashed_indices):
@@ -562,7 +575,7 @@ def test_duplicate_attester_slashing(spec, state):
     yield 'post', None
 
 
-# All AttesterSlashing tests should be adopted for Phase 1 but helper support is not yet there
+# TODO All AttesterSlashing tests should be adopted for SHARDING and later but helper support is not yet there
 
 @with_all_phases
 @spec_state_test
@@ -744,7 +757,8 @@ def test_deposit_top_up(spec, state):
     initial_balances_len = len(state.balances)
     validator_pre_balance = get_balance(state, validator_index)
 
-    yield 'pre', state
+    pre_state = state.copy()
+    yield 'pre', pre_state
 
     block = build_empty_block_for_next_slot(spec, state)
     block.body.deposits.append(deposit)
@@ -756,7 +770,23 @@ def test_deposit_top_up(spec, state):
 
     assert len(state.validators) == initial_registry_len
     assert len(state.balances) == initial_balances_len
-    assert get_balance(state, validator_index) == validator_pre_balance + amount
+
+    # Altair introduces sync committee (sm) reward and penalty
+    sync_committee_reward = sync_committee_penalty = 0
+    if is_post_altair(spec):
+        committee_indices = compute_committee_indices(spec, state, state.current_sync_committee)
+        committee_bits = block.body.sync_aggregate.sync_committee_bits
+        sync_committee_reward, sync_committee_penalty = compute_sync_committee_participant_reward_and_penalty(
+            spec,
+            pre_state,
+            validator_index,
+            committee_indices,
+            committee_bits,
+        )
+
+    assert get_balance(state, validator_index) == (
+        validator_pre_balance + amount + sync_committee_reward - sync_committee_penalty
+    )
 
 
 @with_all_phases
@@ -769,26 +799,24 @@ def test_attestation(spec, state):
     attestation_block = build_empty_block(spec, state, state.slot + spec.MIN_ATTESTATION_INCLUSION_DELAY)
 
     index = 0
-    if spec.fork == PHASE1:
-        shard = spec.compute_shard_from_committee_index(state, index, state.slot)
-        shard_transition = get_shard_transition_of_committee(spec, state, index)
-        attestation_block.body.shard_transitions[shard] = shard_transition
-    else:
-        shard_transition = None
+    # if spec.fork == SHARDING:
+    #     TODO add shard data to block to vote on
 
-    attestation = get_valid_attestation(
-        spec, state, shard_transition=shard_transition, index=index, signed=True, on_time=True
-    )
+    attestation = get_valid_attestation(spec, state, index=index, signed=True)
+
+    if not is_post_altair(spec):
+        pre_current_attestations_len = len(state.current_epoch_attestations)
 
     # Add to state via block transition
-    pre_current_attestations_len = len(state.current_epoch_attestations)
     attestation_block.body.attestations.append(attestation)
     signed_attestation_block = state_transition_and_sign_block(spec, state, attestation_block)
 
-    assert len(state.current_epoch_attestations) == pre_current_attestations_len + 1
-
-    # Epoch transition should move to previous_epoch_attestations
-    pre_current_attestations_root = spec.hash_tree_root(state.current_epoch_attestations)
+    if not is_post_altair(spec):
+        assert len(state.current_epoch_attestations) == pre_current_attestations_len + 1
+        # Epoch transition should move to previous_epoch_attestations
+        pre_current_attestations_root = spec.hash_tree_root(state.current_epoch_attestations)
+    else:
+        pre_current_epoch_participation_root = spec.hash_tree_root(state.current_epoch_participation)
 
     epoch_block = build_empty_block(spec, state, state.slot + spec.SLOTS_PER_EPOCH)
     signed_epoch_block = state_transition_and_sign_block(spec, state, epoch_block)
@@ -796,13 +824,19 @@ def test_attestation(spec, state):
     yield 'blocks', [signed_attestation_block, signed_epoch_block]
     yield 'post', state
 
-    assert len(state.current_epoch_attestations) == 0
-    assert spec.hash_tree_root(state.previous_epoch_attestations) == pre_current_attestations_root
+    if not is_post_altair(spec):
+        assert len(state.current_epoch_attestations) == 0
+        assert spec.hash_tree_root(state.previous_epoch_attestations) == pre_current_attestations_root
+    else:
+        for index in range(len(state.validators)):
+            assert state.current_epoch_participation[index] == spec.ParticipationFlags(0b0000_0000)
+        assert spec.hash_tree_root(state.previous_epoch_participation) == pre_current_epoch_participation_root
 
 
-# In phase1 a committee is computed for SHARD_COMMITTEE_PERIOD slots ago,
+# After SHARDING is enabled, a committee is computed for SHARD_COMMITTEE_PERIOD slots ago,
 # exceeding the minimal-config randao mixes memory size.
 # Applies to all voluntary-exit sanity block tests.
+# TODO: when integrating SHARDING tests, voluntary-exit tests may need to change.
 
 @with_all_phases
 @spec_state_test
@@ -811,7 +845,7 @@ def test_voluntary_exit(spec, state):
     validator_index = spec.get_active_validator_indices(state, spec.get_current_epoch(state))[-1]
 
     # move state forward SHARD_COMMITTEE_PERIOD epochs to allow for exit
-    state.slot += spec.SHARD_COMMITTEE_PERIOD * spec.SLOTS_PER_EPOCH
+    state.slot += spec.config.SHARD_COMMITTEE_PERIOD * spec.SLOTS_PER_EPOCH
 
     signed_exits = prepare_signed_exits(spec, state, [validator_index])
     yield 'pre', state
@@ -839,7 +873,7 @@ def test_double_validator_exit_same_block(spec, state):
     validator_index = spec.get_active_validator_indices(state, spec.get_current_epoch(state))[-1]
 
     # move state forward SHARD_COMMITTEE_PERIOD epochs to allow for exit
-    state.slot += spec.SHARD_COMMITTEE_PERIOD * spec.SLOTS_PER_EPOCH
+    state.slot += spec.config.SHARD_COMMITTEE_PERIOD * spec.SLOTS_PER_EPOCH
 
     # Same index tries to exit twice, but should only be able to do so once.
     signed_exits = prepare_signed_exits(spec, state, [validator_index, validator_index])
@@ -863,7 +897,7 @@ def test_multiple_different_validator_exits_same_block(spec, state):
         for i in range(3)
     ]
     # move state forward SHARD_COMMITTEE_PERIOD epochs to allow for exit
-    state.slot += spec.SHARD_COMMITTEE_PERIOD * spec.SLOTS_PER_EPOCH
+    state.slot += spec.config.SHARD_COMMITTEE_PERIOD * spec.SLOTS_PER_EPOCH
 
     signed_exits = prepare_signed_exits(spec, state, validator_indices)
     yield 'pre', state
@@ -913,7 +947,7 @@ def test_balance_driven_status_transitions(spec, state):
     assert state.validators[validator_index].exit_epoch == spec.FAR_FUTURE_EPOCH
 
     # set validator balance to below ejection threshold
-    state.validators[validator_index].effective_balance = spec.EJECTION_BALANCE
+    state.validators[validator_index].effective_balance = spec.config.EJECTION_BALANCE
 
     yield 'pre', state
 
@@ -927,8 +961,11 @@ def test_balance_driven_status_transitions(spec, state):
     assert state.validators[validator_index].exit_epoch < spec.FAR_FUTURE_EPOCH
 
 
+# Requires always_bls because historical root period and sync committee period is same length
+# so this epoch transition also computes new sync committees which requires aggregation
 @with_all_phases
 @spec_state_test
+@always_bls
 def test_historical_batch(spec, state):
     state.slot += spec.SLOTS_PER_HISTORICAL_ROOT - (state.slot % spec.SLOTS_PER_HISTORICAL_ROOT) - 1
     pre_historical_roots_len = len(state.historical_roots)
@@ -947,12 +984,9 @@ def test_historical_batch(spec, state):
 
 
 @with_all_phases
+@with_presets([MINIMAL], reason="suffices to test eth1 data voting without long voting period")
 @spec_state_test
 def test_eth1_data_votes_consensus(spec, state):
-    if spec.EPOCHS_PER_ETH1_VOTING_PERIOD > 2:
-        return dump_skipping_message("Skip test if config with longer `EPOCHS_PER_ETH1_VOTING_PERIOD` for saving time."
-                                     " Minimal config suffice to cover the target-of-test.")
-
     voting_period_slots = spec.EPOCHS_PER_ETH1_VOTING_PERIOD * spec.SLOTS_PER_EPOCH
 
     offset_block = build_empty_block(spec, state, slot=voting_period_slots - 1)
@@ -991,12 +1025,9 @@ def test_eth1_data_votes_consensus(spec, state):
 
 
 @with_all_phases
+@with_presets([MINIMAL], reason="suffices to test eth1 data voting without long voting period")
 @spec_state_test
 def test_eth1_data_votes_no_consensus(spec, state):
-    if spec.EPOCHS_PER_ETH1_VOTING_PERIOD > 2:
-        return dump_skipping_message("Skip test if config with longer `EPOCHS_PER_ETH1_VOTING_PERIOD` for saving time."
-                                     " Minimal config suffice to cover the target-of-test.")
-
     voting_period_slots = spec.EPOCHS_PER_ETH1_VOTING_PERIOD * spec.SLOTS_PER_EPOCH
 
     pre_eth1_hash = state.eth1_data.block_hash
@@ -1024,25 +1055,25 @@ def test_eth1_data_votes_no_consensus(spec, state):
     yield 'post', state
 
 
-@with_phases([PHASE0])
+@with_all_phases
 @spec_state_test
 def test_full_random_operations_0(spec, state):
     yield from run_test_full_random_operations(spec, state, rng=Random(2020))
 
 
-@with_phases([PHASE0])
+@with_all_phases
 @spec_state_test
 def test_full_random_operations_1(spec, state):
     yield from run_test_full_random_operations(spec, state, rng=Random(2021))
 
 
-@with_phases([PHASE0])
+@with_all_phases
 @spec_state_test
 def test_full_random_operations_2(spec, state):
     yield from run_test_full_random_operations(spec, state, rng=Random(2022))
 
 
-@with_phases([PHASE0])
+@with_all_phases
 @spec_state_test
 def test_full_random_operations_3(spec, state):
     yield from run_test_full_random_operations(spec, state, rng=Random(2023))
